/*  Linux word  The Japanese word processor
 *  fileio.c
 *
 *  Copyright (C) 1999, 2000 MITUIWA Yukio, YOSHIDA Daisuke
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <gtk/gtk.h>

#include "config.h"

#include "fileio.h"
#include "gtkwrap.h"
#include "edit.h"
#include "word.h"

/*****************************************
 * ƼΥե򳫤
 *****************************************/
void open_dialog(gint type) {
  static GtkWidget *dialog;
  gchar title[][64] = {"ե򳫤...",
                      "ե̾դ¸...",
                      "ʸPostScriptǽ...",
                      "ƥȥեɤ߹...",
                      "ƥȥե...",
                      "ʸTeXǽ...",
		       "ƥȤ..."};
  gchar name[][64] =
  {"", "noname.wd1", "output.ps", "", "", "output.tex", "output.txt"};
  static GtkSignalFunc func[] =
  {read_document, save_document, print_document,
   readtext_document, inserttext_document, tex_document, plaintext_document};

  dialog = gtk_file_selection_new (title[type]);
  gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
		     (GtkSignalFunc)gtk_grab_remove, NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(dialog)->ok_button),
		     "clicked",
                     func[type], (gpointer)dialog);
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(dialog)->cancel_button),
                            "clicked", (GtkSignalFunc)gtk_widget_destroy, (gpointer)dialog);
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(dialog), name[type]);
  gtk_widget_show(dialog);
  gtk_grab_add(dialog);
}

/*****************************************
 * ֳ
 *****************************************/
void read_document(GtkWidget *widget, GtkWidget *dialog) {
  read_document_from_name
    (gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog)));
  gtk_widget_destroy(dialog);
}
void read_document_from_name (gchar *filename) {
  gchar buffer[1024];
  gint  i, j, x, y, table[2][MAX_FONT];
  FILE  *fp;
  GdkFont *font;

  unref_all_pixmaps();
  cpage = disptop = 0;
  if((fp = fopen(filename, "r")) == NULL) return;
  fscanf(fp, "%s\n", buffer);
  if(strcmp(buffer, "#!Word-document-file.") != 0) {
    fclose(fp);
    return;
  }
  fscanf(fp, "%s", buffer);
  j = 0;
  while(!feof(fp) && buffer[0] != '*') {
    table[KANJI_FONT][j] = 0;
    for(i = 0;fontnames[KANJI_FONT][i][0] != '0';i++) {
      if(strcmp(buffer, fontnames[KANJI_FONT][i]) == 0) {
        table[KANJI_FONT][j] = i;
        break;
      }
    }
    fscanf(fp, "%s", buffer);
    j++;
  }
  fscanf(fp, "%s", buffer);
  j = 0;
  while(!feof(fp) && buffer[0] != '*') {
    table[ASCII_FONT][j] = 0;
    for(i = 0;fontnames[ASCII_FONT][i][0] != '0';i++) {
      if(strcmp(buffer, fontnames[ASCII_FONT][i]) == 0) {
        table[ASCII_FONT][j] = i;
        break;
      }
    }
    fscanf(fp, "%s", buffer);
    j++;
  }
  fscanf(fp, "%d", &paper_width);
  fscanf(fp, "%d", &paper_height);
  fscanf(fp, "%d", &paper_top);
  fscanf(fp, "%d", &paper_bottom);
  fscanf(fp, "%d", &paper_left);
  fscanf(fp, "%d", &paper_right);
  page_x = paper_width - paper_left - paper_right;
  page_y = paper_height - paper_top - paper_bottom;
  max_dsp_num = page_y / 7;
  pix_width = page_x * 2;
  fscanf(fp, "%s", buffer);
  while(!feof(fp) && buffer[0] != '*') {
    fscanf(fp, "%s", buffer);
  }
  x = y = 0;
  fscanf(fp, "%x", &(lines[y].attr));
  fscanf(fp, "%d", &(lines[y].vspace));
  fscanf(fp, "%d", &(lines[y].lefttab));
  fscanf(fp, "%d", &(lines[y].righttab));
  fscanf(fp, "%d", &(lines[y].firsttab));
  fscanf(fp, "%d", &(lines[y].reserved1));
  fscanf(fp, "%d", &(lines[y].reserved2));
  fscanf(fp, "%d", &(lines[y].reserved3));
  fscanf(fp, "%d", &(lines[y].reserved4));
  if(lines[y].pixmap != (GdkPixmap*)NULL) {
    gdk_pixmap_unref(lines[y].pixmap);
    lines[y].pixmap = (GdkPixmap*)NULL;
  }
  while(!feof(fp)) {
    fscanf(fp, "%x", &(master[y][x].code));
    fscanf(fp, "%d", &(master[y][x].point));
    fscanf(fp, "%x", &(master[y][x].attr));
    fscanf(fp, "%d", &i);
    if((master[y][x].code & 0xff00) != 0xff00) {
      if(master[y][x].code > 0xff)
        master[y][x].findex = table[KANJI_FONT][i];
      else
        master[y][x].findex = table[ASCII_FONT][i];
      font = get_fontset(master[y][x]);
      master[y][x].width = gdk_text_width_wc(font, &master[y][x].code, 1);
    } else {
      master[y][x].width = 0;
    }
    if(master[y][x].code == GDK_Return || master[y][x].attr & WordChr_CONT) {
      draw_line_pixmap(y);
      fscanf(fp, "%x", &(lines[++y].attr));
      fscanf(fp, "%d", &(lines[y].vspace));
      fscanf(fp, "%d", &(lines[y].lefttab));
      fscanf(fp, "%d", &(lines[y].righttab));
      fscanf(fp, "%d", &(lines[y].firsttab));
      fscanf(fp, "%d", &(lines[y].reserved1));
      fscanf(fp, "%d", &(lines[y].reserved2));
      fscanf(fp, "%d", &(lines[y].reserved3));
      fscanf(fp, "%d", &(lines[y].reserved4));
      if(lines[y].pixmap != (GdkPixmap*)NULL) {
        gdk_pixmap_unref(lines[y].pixmap);
        lines[y].pixmap = (GdkPixmap*)NULL;
      }
      x = -1;
    }
    x++;
  }
  master_.x = master_.y = 0;
  fclose(fp); 
  tab_draw(tab_set, 0, TAB_DRAW);
  draw_etext();
}

/*****************************************
 * ƥȤ򥤥ݡ
 *****************************************/
void readtext_document(GtkWidget *widget, GtkWidget *dialog) {
  gchar *filename;
  gint i, c, size, line;
  gchar buf[9];
  GdkWChar wcs[9];
  FILE *fp;

  unref_all_pixmaps();
  cpage = disptop = 0;
  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog));
  if((fp = fopen(filename, "r")) == NULL) return;
  set_default_attr();
  lines[0].attr = LineAttr_NP | LineAttr_EOF;
  lines[0].lefttab = 0;
  lines[0].righttab = 0;
  master_.x = master_.y = 0;
  master[0][0].attr = editbuf_attr;
  master[0][0].point = editbuf_point;
  master[0][0].findex = editbuf_findex[ASCII_FONT];
  master[0][0].code = GDK_Return;
  master[0][0].width = 0;
  editbuf_new();
  c = fgetc(fp);
  line = 0;
  while(!feof(fp)) {
    buf[0] = (gchar)c;
    if(c == 0x0a) {
      wcs[0] = GDK_Return;
      wcs[1] = 0;
      set_wcs_editbuf(wcs);
      master_ = insert_editbuf(master_);
      editbuf_new();
      c = fgetc(fp);
      continue;
    }
    if(c & 0x80) {
      buf[1] = (gchar)fgetc(fp);
      buf[2] = '\0';
    } else if(c >= 0x20) {
      buf[1] = '\0';
    } else if(c == '\t') {
      for(i = 0;i < 8;i++) buf[i] = ' ';
      buf[8] = '\0';
    } else {
      c = fgetc(fp);
      continue;
    }
    size = gdk_mbstowcs (wcs, buf, 8);
    wcs[size] = 0;
    set_wcs_editbuf(wcs);
    c = fgetc(fp);
  }
  tab_draw(tab_set, 0, TAB_DRAW);
  master_.x = master_.y = 0;
  draw_etext();
  fclose(fp); 
  gtk_widget_destroy(dialog);
}

/*****************************************
 * ƥȤ
 *****************************************/
void inserttext_document(GtkWidget *widget, GtkWidget *dialog) {
  gchar *filename;
  gint i, c, size;
  gchar buf[9];
  GdkWChar wcs[9];
  FILE *fp;

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog));
  if((fp = fopen(filename, "r")) == NULL) return;
  set_default_attr();
  editbuf_new();
  c = fgetc(fp);
  while(!feof(fp)) {
    buf[0] = (gchar)c;
    if(c == 0x0a) {
      wcs[0] = GDK_Return;
      wcs[1] = 0;
      set_wcs_editbuf(wcs);
      master_ = insert_editbuf(master_);
      editbuf_new();
      c = fgetc(fp);
      continue;
    }
    if(c & 0x80) {
      buf[1] = (gchar)fgetc(fp);
      buf[2] = '\0';
    } else if(c >= 0x20) {
      buf[1] = '\0';
    } else if(c == '\t') {
      for(i = 0;i < 8;i++) buf[i] = ' ';
      buf[8] = '\0';
    } else {
      c = fgetc(fp);
      continue;
    }
    size = gdk_mbstowcs (wcs, buf, 8);
    wcs[size] = 0;
    set_wcs_editbuf(wcs);
    c = fgetc(fp);
  }
  draw_etext();
  fclose(fp); 
  gtk_widget_destroy(dialog);
}

/*****************************************
 * ¸
 *****************************************/
void save_document(GtkWidget *widget, GtkWidget *dialog) {
  gchar *filename;
  gint  i, x, y, xe, ye;
  FILE  *fp;

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog));
  if((fp = fopen(filename, "w")) == NULL) return;
  fprintf(fp, "#!Word-document-file.\n");
  for(i = 0;fontnames[KANJI_FONT][i][0] != '0';i++) fprintf(fp, "%s  ", fontnames[KANJI_FONT][i]);
  fprintf(fp, "*  ");
  for(i = 0;fontnames[ASCII_FONT][i][0] != '0';i++) fprintf(fp, "%s  ", fontnames[ASCII_FONT][i]);
  fprintf(fp, "*  ");
  fprintf(fp, "%d %d %d %d %d %d *\n", paper_width, paper_height, 
          paper_top, paper_bottom, paper_left, paper_right);
  ye = get_end_page();
  for(y = 0;y <= ye;y++) {
    xe = get_end_line(y);
    fprintf(fp, "%04x ", lines[y].attr);
    fprintf(fp, "%d ", lines[y].vspace);
    fprintf(fp, "%d ", get_lefttab(y));
    fprintf(fp, "%d ", lines[y].righttab);
    fprintf(fp, "%d ", lines[y].firsttab);
    fprintf(fp, "%d ", lines[y].reserved1);
    fprintf(fp, "%d ", lines[y].reserved2);
    fprintf(fp, "%d ", lines[y].reserved3);
    fprintf(fp, "%d ", lines[y].reserved4);
    for(x = 0;x <= xe;x++){
      fprintf(fp, "%x ", (gint)master[y][x].code);
      fprintf(fp, "%d ", master[y][x].point);
      fprintf(fp, "%02x ", master[y][x].attr);
      fprintf(fp, "%d ", master[y][x].findex);
    }
    fprintf(fp, "\n");
  }
  fclose(fp);
  gtk_widget_destroy(dialog);
}

/*****************************************
 * PostScript
 *****************************************/
void print_document(GtkWidget *widget, GtkWidget *dialog) {
  gchar *filename;
  FILE  *fp;

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog));
  if((fp = fopen(filename, "w")) == NULL) return;
  psout(fp);
  fclose(fp);
  gtk_widget_destroy(dialog);
}

/*****************************************
 * ְ
 *****************************************/
void lpr_document(GtkWidget *widget, GtkWidget **widgets) {
  GtkWidget *dialog, *entry;
  FILE *tmpfile;

  dialog = widgets[0];
  entry = widgets[1];

  if((tmpfile = popen(gtk_entry_get_text(GTK_ENTRY(entry)), "w")) == NULL)
    return;
  psout(tmpfile);
  pclose(tmpfile);

  gtk_widget_destroy(dialog);
}

/*****************************************
 * ֥ץӥ塼
 *****************************************/
void preview_document(GtkWidget *widget, GtkWidget **widgets) {
  GtkWidget *dialog, *entry;
  FILE *tmpfile;
  gchar *preview_command;
  gint preview_command_length;

  if((tmpfile = fopen(PREVIEWTMP, "w")) == NULL) return;
  psout(tmpfile);
  fclose(tmpfile);

  dialog = widgets[0];
  entry = widgets[1];

  preview_command_length =
    strlen(gtk_entry_get_text(GTK_ENTRY(entry))) + strlen(PREVIEWTMP) + 3;
  preview_command = g_malloc(preview_command_length);
  sprintf (preview_command, "%s %s",
	   gtk_entry_get_text(GTK_ENTRY(entry)), PREVIEWTMP);
  if((tmpfile = popen(preview_command, "r")) == NULL)
    return;
  pclose(tmpfile);
  g_free(preview_command);

  gtk_widget_destroy(dialog);
}

/*****************************************
 * TeXνϡ
 *****************************************/
void tex_document(GtkWidget *widget, GtkWidget *dialog) {
  gchar *filename;
  FILE  *fp;

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog));
  if((fp = fopen(filename, "w")) == NULL) return;
  texout(fp);
  fclose(fp);
  gtk_widget_destroy(dialog);
}

/*****************************************
 * ƥȽ
 *****************************************/
void plaintext_document(GtkWidget *widget, GtkWidget *dialog) {
  gchar *filename;
  FILE  *fp;

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(dialog));
  if((fp = fopen(filename, "w")) == NULL) return;
  plaintextout(fp);
  fclose(fp);
  gtk_widget_destroy(dialog);
}

/*****************************************
 * PostScript
 *****************************************/
void psout(FILE *fp) {
  gint x, y, mx, my, height, c, point, i, b, curpage, cp;
  GdkWChar wc[2];
  gchar *mc;

  fprintf(fp, "%%!PS-Adobe\n");
  fprintf(fp, "%d %d moveto\n", paper_left, paper_height - paper_top);
/*  fprintf(fp, "\n%%%%Page: 1 1\n\n");*/
  my = get_end_page();
  wc[1] = 0;
  curpage = 0;
  for(y = 0;y <= my;y++) {
    if(get_page_top(curpage + 1, &cp) == y && y > 0) {
      curpage++;
/*      fprintf(fp, "\n%% %% Page: %d %d\n\n", curpage + 1, curpage + 1);*/
      fprintf(fp, "showpage\n");
      fprintf(fp, "%d %d moveto\n", paper_left, paper_height - paper_top);
    }
    mx = get_end_line(y);
    height = get_line_height(y);
    fprintf(fp, "currentpoint exch pop %d sub %d exch moveto\n", 
            height + height / 6 + lines[y].vspace, get_pix_x(y, 0) + paper_left);
    for(x = 0;x <= mx;x++) {
      wc[0] = master[y][x].code;
      point = master[y][x].point;
      if(master[y][x].attr & WordChr_UL) {
        fprintf(fp, "currentpoint %d sub\n", height / 6);
      }
      if(wc[0] > 0xff && (wc[0] & 0xff00) != 0xff00) {
        mc = gdk_wcstombs(wc);
        c = (gint)mc[0] & 0xff;
        c <<= 8;
        c |= (gint)mc[1] & 0xff;
        if(master[y][x].attr & WordChr_IT) {
          fprintf(fp, "%s-EUC-H findfont [%d 0 %d %d 0 0] makefont setfont\n",
                  psfontnames[KANJI_FONT][master[y][x].findex], point, point / 4, point);
        } else {
          fprintf(fp, "%s-EUC-H findfont %d scalefont setfont\n",
                  psfontnames[KANJI_FONT][master[y][x].findex], point);
       }
        if(master[y][x].attr & WordChr_BOLD) {
          b = master[y][x].width / 36 + 1;
          for(i = 1;i <= b;i++) {
            fprintf(fp, "currentpoint\n");
            fprintf(fp, "<%04x> show\n",c);
            fprintf(fp, "exch %d add exch moveto\n", i);
          }
        }
        fprintf(fp, "<%04x> show\n",c);
      } else if(wc[0] >= 0x20 && wc[0] <= 0x7f) {
        c = (gint)wc[0] & 0xff;
        if(master[y][x].attr & WordChr_IT) {
          fprintf(fp, "%s findfont [%d 0 %d %d 0 0] makefont setfont\n",
                  psfontnames[ASCII_FONT][master[y][x].findex], point, point / 2, point);
        } else {
          fprintf(fp, "%s findfont %d scalefont setfont\n", 
                  psfontnames[ASCII_FONT][master[y][x].findex], point);
        }
        if(master[y][x].attr & WordChr_BOLD) {
          b = master[y][x].width / 36 + 1;
          for(i = 1;i <= b;i++) {
            fprintf(fp, "currentpoint\n");
            if(c == '(' || c == ')' || c == '\\')
              fprintf(fp, "(\\%c) show\n", c);
            else
              fprintf(fp, "(%c) show\n", c);
            fprintf(fp, "exch %d add exch moveto\n", i);
          }
        }
        if(c == '(' || c == ')' || c == '\\')
          fprintf(fp, "(\\%c) show\n", c);
        else
          fprintf(fp, "(%c) show\n", c);
      }
      if(master[y][x].attr & WordChr_UL) {
        fprintf(fp, "gsave 0 %d rmoveto lineto stroke grestore\n", (height / 6) * (-1));
      }
    }
  }
  fprintf(fp, "showpage\n");
/*  fprintf(fp, "%%%%Trailer\n%%Pages: %d\n%%%%EOF\n", curpage + 1);*/
}

/*****************************************
 * ƥȥե
 *****************************************/
void plaintextout(FILE *fp) {
  gint x, y, mx, my, curpage, cp;
  GdkWChar wc[2];
  gchar *mc;

  my = get_end_page();
  wc[1] = 0;
  curpage = 0;
  for(y = 0;y <= my;y++) {
    if(get_page_top(curpage + 1, &cp) == y && y > 0) {
      curpage++;
    }
    mx = get_end_line(y);
    for(x = 0;x <= mx;x++) {
      wc[0] = master[y][x].code;
      if(wc[0] > 0xff && (wc[0] & 0xff00) != 0xff00) {
        mc = gdk_wcstombs(wc);
	fprintf(fp, "%s", mc);
      } else if(wc[0] >= 0x20 && wc[0] <= 0x7f) {
	fprintf(fp, "%c", wc[0]);
      } else if (wc[0] == GDK_Return) {
	fprintf (fp, "\n");
      }
    }
  }
}

/*****************************************
 * TeXե
 *****************************************/
void texout(FILE *fp) {
  gint x, y, mx, my, lineattr, point, ul, save1, save2;
  GdkWChar wc[2];
  gchar *mc;

  fprintf(fp, "\\documentstyle[a4j,12pt,epsf]{jarticle}\n");
  fprintf(fp, "\\pagestyle{empty}\n");
  fprintf(fp, "\\begin{document}\n");
  ul = 0;
  save1 = 12;
  lineattr = lines[0].attr;
  switch(lineattr & LineAttr_Form) {
  case LineAttr_Center:
    fprintf(fp, "\\begin{center} ");
    break;
  case LineAttr_Right:
    fprintf(fp, "\\begin{flushright} ");
    break;
  }
  my = get_end_page();
  wc[1] = 0;
  for(y = 0;y <= my;y++) {
    mx = get_end_line(y);
    for(x = 0;x <= mx;x++) {
      wc[0] = master[y][x].code;
      point = master[y][x].point;
      if(master[y][x].attr & WordChr_UL) {
        if(ul == 0) {
          ul = 1;
          fprintf(fp, "\\underline{");
          save2 = point;
        }
      } else {
      	if(ul == 1) {
          ul = 0;
          fprintf(fp, "}");
          tex_size(NULL, save2);
        }
      } 
      if(wc[0] > 0xff && (wc[0] & 0xff00) != 0xff00) {
        tex_size(fp, point);
        mc = gdk_wcstombs(wc);
        fprintf(fp, "%c%c", mc[0], mc[1]);
      } else if(wc[0] >= 0x20 && wc[0] <= 0x7f) {
        tex_size(fp, point);
        switch(wc[0]) {
        case ' ' :
        case '$' :
        case '{' :
        case '}' :
        case '&' :
        case '|' :
          fprintf(fp, "\\%c", wc[0]);
          break;
        case '\\' :
          fprintf(fp, "\\\\");
          break;
        default:
          fprintf(fp, "%c", wc[0]);
        }
     } else if(wc[0] == GDK_Return) {
      	if(ul == 1) {
          ul = 0;
          fprintf(fp, "}");
          tex_size(NULL, save2);
        }
        switch(lineattr & LineAttr_Form) {
        case LineAttr_Center:
          tex_size(NULL, save1);
          fprintf(fp, "\\end{center}\n");
          break;
        case LineAttr_Right:
          tex_size(NULL, save1);
          fprintf(fp, "\\end{flushright}\n");
          break;
        default:
          fprintf(fp, "\n\n");
        }
        lineattr = lines[y + 1].attr;
        if(y < my) {
          save1 = point;
          switch(lineattr & LineAttr_Form) {
          case LineAttr_Center:
            fprintf(fp, "\\begin{center} ");
            break;
          case LineAttr_Right:
            fprintf(fp, "\\begin{flushright} ");
            break;
          }
        }
      }
    }
  }
  if(ul == 1) {
    ul = 0;
    fprintf(fp, "}");
    tex_size(NULL, save2);
  }
  fprintf(fp, "\\end{document}\n");
}

/*****************************************
 * TeXΥ̿
 *****************************************/
void tex_size(FILE *fp, gint point) {
  static gchar size[32] = "\\normalsize ";
  gchar cursize[32];

  if(point < 8) strcpy(cursize, "\\tiny ");
  else if(point < 10) strcpy(cursize, "\\scriptsize ");
  else if(point < 11) strcpy(cursize, "\\footnotesize ");
  else if(point < 12) strcpy(cursize, "\\small ");
  else if(point < 14) strcpy(cursize, "\\normalsize ");
  else if(point < 17) strcpy(cursize, "\\large ");
  else if(point < 20) strcpy(cursize, "\\Large ");
  else if(point < 25) strcpy(cursize, "\\LARGE ");
  else if(point < 30) strcpy(cursize, "\\huge ");
  else strcpy(cursize, "\\Huge ");
  if(strcmp(cursize, size) != 0) {
    strcpy(size, cursize);
    if(fp != NULL) fprintf(fp, size);
  }
}

/*****************************************
 * Хåؿã
 *****************************************/
void read_text() {
  open_dialog(READTEXT_DIALOG);
}

void insert_textfile() {
  open_dialog(INSERTTEXT_DIALOG);
}

void read_word() {
  open_dialog(READ_DIALOG);
}

void save_word() {
  open_dialog(SAVE_DIALOG);
}

void print_word() {
  open_dialog(PRINT_DIALOG);
}

void tex_out() {
  open_dialog(TEX_DIALOG);
}

void plaintext_out() {
  open_dialog(PLAIN_DIALOG);
}

void lpr_out() {
  static gchar buffer[32];
  strncpy (buffer, LPRCOMMAND, 30);
  create_dialog("ѤΥޥ", buffer, 30, GTK_SIGNAL_FUNC(lpr_document));
}

void preview_out() {
  gchar buffer[200];
  strncpy (buffer, PSVIEWER, 190);
  create_dialog("ץӥ塼ѤΥޥ", buffer, 190, GTK_SIGNAL_FUNC(preview_document));
}
