/*  Linux word  The Japanese word processor
 *  word.c
 *
 *  Copyright (C) 1999, 2000 MITUIWA Yukio, YOSHIDA Daisuke
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>

#include "word.h"
#include "gtkwrap.h"

gint paper_width = A4WIDTH;
gint paper_left = 20;
gint paper_right = 20;
gint paper_top = 30;
gint paper_bottom = 25;
gint paper_height = A4HEIGHT;
gint vspace = 0;
gint tab1 = 0, tab2 = 0;

void redraw_init() {
  gint min, max;

  unref_all_pixmaps();
  cpage = disptop = 0;
  page_x = paper_width - paper_left - paper_right;
  page_y = paper_height - paper_top - paper_bottom;
  max_dsp_num = page_y / 7;
  pix_width = page_x * 2;
  max = -1;
  while(get_lines_info(max + 1, &min, &max)) {
    max = ajust_master(min, max);
  }
  master_.y = master_.x = 0;
  set_page(0);
  replace_tab(tab_set);
}

gint blink_cursor() {
  static gint count = 2, pix_x, pix_y, height;

  if(count > 3) {
    count = 0;
    pix_y = get_pix_y(master_.y) + lines[master_.y].vspace;
    pix_x = get_pix_x(master_.y, master_.x); 
    height = get_line_height(master_.y);
  }
  if(count <= 2) gdk_draw_rectangle(etext->window, etext->style->black_gc,
                                    TRUE, pix_x, pix_y, 1, height);
  if(count > 2) gdk_draw_rectangle(etext->window, etext->style->white_gc, 
                                   TRUE, pix_x, pix_y, 1, height);
  count++;
  return 1;
}

void draw_line_pixmap(gint y) {
  gint i, b, x, x_pix, s, width, point;
  GdkWChar code;
  GdkFont *font;
  GdkGC   *gc;

  if(y < disptop || y >= (disptop + max_dsp_num)) return; 
  if(lines[y].pixmap == NULL)
    lines[y].pixmap = gdk_pixmap_new(etext->window, pix_width, pix_height, -1);
  gdk_draw_rectangle(lines[y].pixmap, etext->style->white_gc, TRUE, 0, 0, pix_width, pix_height);
  s = get_end_line(y);
  point = get_line_height(y);
  x_pix = 0;
  for(x = 0;x <= s;x++) {
    if(master[y][x].code != GDK_Return) {
      code = master[y][x].code;
      font = get_fontset(master[y][x]);
      width = master[y][x].width;
      gc = etext->style->black_gc;
      if(master[y][x].attr & WordChr_REV) {
        gc = etext->style->white_gc;
        gdk_draw_rectangle(lines[y].pixmap, etext->style->black_gc,
                           TRUE, x_pix, 0, width, pix_height - 1);
      }
      gdk_draw_text_wc(lines[y].pixmap, font, gc, x_pix, pix_height - 1 - (point / 6), &code, 1);
      if(master[y][x].attr & WordChr_BOLD) {
      	b = width / 36 + 1;
      	for(i = 1;i <= b;i++)
          gdk_draw_text_wc(lines[y].pixmap, font, gc, x_pix + i, pix_height - 1 - (point / 6), &code, 1);  
      }
      if(master[y][x].attr & WordChr_UL)
        gdk_draw_line(lines[y].pixmap, gc, x_pix, pix_height - 1, x_pix + width, pix_height - 1);
      x_pix += width;
    } else {
      gc = etext->style->black_gc;
      if(master[y][x].attr & WordChr_REV) {
        gc = etext->style->white_gc;
        gdk_draw_rectangle(lines[y].pixmap, etext->style->black_gc, TRUE, x_pix, 0, 16, pix_height - 1);
      }
      gdk_draw_rectangle(lines[y].pixmap, gc, TRUE, x_pix + 2, pix_height - 1 - (point / 6) - 5, 11, 3);
      gdk_draw_rectangle(lines[y].pixmap, gc, TRUE, x_pix + 10, 0, 3, pix_height - 1 - (point / 6) - 5);
      gdk_draw_line(lines[y].pixmap, gc,
                    x_pix + 2, pix_height - 1 - (point / 6) - 5, x_pix + 5, pix_height - 1 - (point / 6) - 8);
      gdk_draw_line(lines[y].pixmap, gc,
                    x_pix + 2, pix_height - 1 - (point / 6) - 3, x_pix + 5, pix_height - 1 - (point / 6));
    }
  }
  set_input_dialog();
}

void draw_etext() {
  gint s, i, pix_y, height, width;
  inc_pixy value;

  if((get_pix_y(master_.y) + get_line_height(master_.y)) >= page_y) set_page_notdraw(cpage + 1);
  if(master_.y < disptop) set_page_notdraw(cpage - 1);
  gdk_draw_rectangle(etext->window, etext->style->white_gc, TRUE, 0, 0, page_x, page_y);
  s = get_end_page();
  pix_y = inc_line_pix_y(INC_INIT, &value);
  for(i = disptop;i <= s && i < (disptop + max_dsp_num);i++) {
    height = get_line_height(i);
    if(lines[i].pixmap) {
      gdk_draw_pixmap(etext->window,etext->style->fg_gc[GTK_WIDGET_STATE (etext)],
                      lines[i].pixmap, 0, pix_height - height / 6 - height, get_pix_x(i, 0), 
                      pix_y + lines[i].vspace, page_x, height + height / 6);
    }
    pix_y = inc_line_pix_y(INC_COUNT, &value);
  }
  gdk_window_get_size (etext->window, &width, &height);
  gdk_draw_rectangle(etext->window, etext->style->white_gc, TRUE, page_x, 0, width, height);
  gdk_draw_rectangle(etext->window, etext->style->white_gc, TRUE, 0, page_y, width, height);
  gdk_draw_line(etext->window, etext->style->black_gc, page_x, 0, page_x, page_y);
  gdk_draw_line(etext->window, etext->style->black_gc, page_x, page_y, 0, page_y);
}

gint get_pix_x(gint y, gint x) {
  gint pix_x, max_x, i;

  i = max_x = 0;
  while(end_line_query(y, i) == FALSE && i < MAX_COLUMN - 1) {
    max_x += master[y][i].width;
    i++;
  }
  max_x += master[y][i].width;
  i = pix_x = 0;
  while(i < x && end_line_query(y, i) == FALSE && i < MAX_COLUMN - 1) {
    pix_x += master[y][i].width;
    i++;
  }
  switch(lines[y].attr & LineAttr_Form) {
  case LineAttr_Center:
    return (page_x - lines[y].righttab - get_lefttab(y) - max_x) / 2 + pix_x + get_lefttab(y);
  case LineAttr_Right:
    return page_x - lines[y].righttab - max_x + pix_x;
  case LineAttr_Left:
    return pix_x + get_lefttab(y);
  }
  return pix_x;
}

gint get_word_y(gint py) {
  gint pix_y, i;
  inc_pixy value;

  pix_y = inc_line_pix_y(INC_INIT, &value);
  for(i = disptop;!(lines[i].attr & LineAttr_EOF) && (i < MAX_ROW - 1);i++){
    pix_y = inc_line_pix_y(INC_COUNT, &value);
    if(pix_y > py) break;
  }
  return i;
}

gint get_word_x(gint y, gint px) {
  gint pix_x, max_x, i;

  i = max_x = 0;
  while(end_line_query(y, i) == FALSE && i < MAX_COLUMN - 1) {
    max_x += master[y][i].width;
    i++;
  }
  max_x += master[y][i].width;
  switch(lines[y].attr & LineAttr_Form) {
  case LineAttr_Center:
    pix_x = (page_x - lines[y].righttab - get_lefttab(y) - max_x) / 2 + get_lefttab(y);
    break;
  case LineAttr_Right:
    pix_x = page_x - lines[y].righttab - max_x;
    break;
  default:
    pix_x = get_lefttab(y);
  }
  i = 0;
  while(end_line_query(y, i) == FALSE && i < MAX_COLUMN - 1) {
    pix_x += master[y][i].width;
    if(pix_x > px) break;
    i++;
  }
  return i;
}

void create_input_dialog (GtkWidget *widget) {
  if (gdk_im_ready () && (ic_attr = gdk_ic_attr_new ()) != NULL) {
    gint width, height;
    GdkColormap *colormap;
    GdkEventMask mask;
    GdkICAttr *attr = ic_attr;
    GdkICAttributesType attrmask = GDK_IC_ALL_REQ;
    GdkIMStyle style;
    GdkIMStyle supported_style = GDK_IM_PREEDIT_NONE | GDK_IM_PREEDIT_NOTHING |
	                           GDK_IM_PREEDIT_POSITION | GDK_IM_STATUS_NONE |
	                           GDK_IM_STATUS_NOTHING;
    if(widget->style && widget->style->font->type != GDK_FONT_FONTSET)
      supported_style &= ~GDK_IM_PREEDIT_POSITION;
    attr->style = style = gdk_im_decide_style (supported_style);
    attr->client_window = widget->window;
    if((colormap = gtk_widget_get_colormap (widget)) != gtk_widget_get_default_colormap ()) {
      attrmask |= GDK_IC_PREEDIT_COLORMAP;
      attr->preedit_colormap = colormap;
    }
    switch (style & GDK_IM_PREEDIT_MASK) {
    case GDK_IM_PREEDIT_POSITION:
      if(widget->style && widget->style->font->type != GDK_FONT_FONTSET) {
        g_warning ("ʴѴΥեȤǤޤǤ");
        break;
      }
      attrmask |= GDK_IC_PREEDIT_POSITION_REQ;
      gdk_window_get_size (widget->window, &width, &height);
      attr->spot_location.x = 0;
      attr->spot_location.y = 14;
      attr->preedit_area.x = 0;
      attr->preedit_area.y = 0;
      attr->preedit_area.width = width;
      attr->preedit_area.height = height;
      attr->preedit_fontset = widget->style->font;
      break;
    }
    ic = gdk_ic_new (attr, attrmask);
    if(ic == NULL) g_warning ("ʴѴѤǤޤ");
    else {
      mask = gdk_window_get_events (widget->window);
      mask |= gdk_ic_get_events(ic);
      gdk_window_set_events (widget->window, mask);
      gdk_im_begin(ic, widget->window);
    }
  }
}

void set_input_dialog() {
  if(ic && (gdk_ic_get_style(ic) & GDK_IM_PREEDIT_POSITION)) {
    ic_attr->spot_location.x = get_pix_x(master_.y, master_.x);
    ic_attr->spot_location.y = get_pix_y(master_.y) + get_line_height(master_.y);
    ic_attr->spot_location.y += lines[master_.y].vspace;
    gdk_ic_set_attr (ic, ic_attr, GDK_IC_SPOT_LOCATION);
  }
}

gint unref_pixmaps() {
  gint i;

  for(i = disptop;i < (disptop + max_dsp_num);i++) {
    if(lines[i].pixmap != NULL) {
      gdk_pixmap_unref(lines[i].pixmap);
      lines[i].pixmap = NULL;
    }
  }
  return 0;
}

gint unref_all_pixmaps() {
  gint i;

  for(i = 0;i < MAX_ROW;i++) {
    if(lines[i].pixmap != NULL) {
      gdk_pixmap_unref(lines[i].pixmap);
      lines[i].pixmap = NULL;
    }
  }
  return 0;
}

void set_page_notdraw(gint pg){
  gint i, height;
  inc_pixy value;

  unref_pixmaps();
  disptop = get_page_top(pg, &cpage);
  height = inc_line_pix_y(INC_INIT, &value);
  for(i = disptop;!(lines[i].attr & LineAttr_EOF) && (i < MAX_ROW - 1);i++) {
    if(height >= page_y) break;
    height = inc_line_pix_y(INC_COUNT, &value);
    draw_line_pixmap(i);
  }
  draw_line_pixmap(i);
}

void set_page(gint pg){
  gint i, height;
  inc_pixy value;

  unref_pixmaps();
  disptop = get_page_top(pg, &cpage);
  height = inc_line_pix_y(INC_INIT, &value);
  for(i = disptop;!(lines[i].attr & LineAttr_EOF) && (i < MAX_ROW - 1);i++) {
    if(height >= page_y) break;
    height = inc_line_pix_y(INC_COUNT, &value);
    draw_line_pixmap(i);
  }
  draw_line_pixmap(i);
  draw_etext();
}

gint get_lefttab(gint y) {
  gint tab, min, max;

  if(lines[y].attr & LineAttr_Tabular) {
    get_lines_info(y, &min, &max);
    tab = page_x - lines[min - 1].righttab;
  } else {
    tab = lines[y].lefttab;
  }
  return tab;
}

gint ajust_master(gint min, gint max) {
  gint i, cy, cx, mx, pos, incline, width;

  pos = 0;
  for(cy = min;cy <= max;cy++) {
    mx = get_end_line(cy);
    for(cx = 0;cx <= mx;cx++) {
      ajustbuf[pos] = master[cy][cx];
      ajustbuf[pos++].attr &= ~WordChr_CONT;
    }
  }
  incline = max - min;
  pos = cx = width = 0, cy = min;
  do {
    if(width + ajustbuf[pos].width >= (page_x - lines[cy].righttab - get_lefttab(cy)) && cx > 0) {
        master[cy][cx - 1].attr |= WordChr_CONT;
        draw_line_pixmap(cy);
        width = 0;
    	cy++, cx = 0;
    	incline--;
    	if(incline < 0) insert_line(cy);
    }
    width += ajustbuf[pos].width;
    master[cy][cx++] = ajustbuf[pos++];
  } while(master[cy][cx - 1].code != GDK_Return);
  draw_line_pixmap(cy);
  for(i = 0;i < incline;i++) delete_line(cy + 1);
  return cy;
}

Position insert_editbuf(Position current) {
  gint save, ed, y, x, ex, incline, width, bx;
  gint cx, cy;
  Position pos;

  if(editbuf_pos <= 0) return current;
  y = current.y, x = current.x;
  incline = 0;
  cy = y, cx = x;
  ed = editbuf_pos;
  while(master[cy][cx].code != GDK_Return) {
    editbuf[editbuf_pos++] = master[cy][cx];
    editbuf[editbuf_pos - 1].attr &= ~WordChr_CONT;
    if(master[cy][cx].attr & WordChr_CONT) {
      cy++;
      cx = -1;
      incline--;
    }
    cx++;
  }
  editbuf[editbuf_pos++] = master[cy][cx];
  editbuf[editbuf_pos - 1].attr &= ~WordChr_CONT;
  bx = width = 0;
  cy = y;
  while(bx < x) width += master[y][bx++].width;
  save = width, cx = ex = 0;
  while(ex < editbuf_pos) {
    if(editbuf[ex].code == GDK_Return || 
       ((width + editbuf[ex].width) >= (page_x - lines[y].righttab - get_lefttab(y)) && (cx + bx) > 0)) {
      incline++;
      width = cx = bx = 0;
    }
    width += editbuf[ex++].width;
    cx++;
  }
  incline--;
  while(incline < 0) {
    delete_line(y + 1);
    incline++;
  }
  while(incline > 0) {
    insert_line(y + 1);
    incline--;
  }
  ex = 0, cy = y, cx = x, width = save;
  while(ex < editbuf_pos) {
    if(width + editbuf[ex].width >= (page_x - lines[y].righttab - get_lefttab(y)) && cx > 0) {
        master[cy][cx - 1].attr |= WordChr_CONT;
        draw_line_pixmap(cy);
        width = 0;
    	cy++, cx = 0;
    }
    width += editbuf[ex].width;
    master[cy][cx++] = editbuf[ex++];
    if(master[cy][cx - 1].code == GDK_Return) {
      draw_line_pixmap(cy);
      width = 0;
      cy++, cx = 0;
    }
    if(ex == ed){
      pos.x = cx, pos.y = cy;
      /*draw_line_pixmap(cy);*/
    }
  }
  return pos;
}

void set_rev(Position *start, Position *end, gint type) {
  Position c, max;
  gint size, state, flag;
  GdkFont *font;
  gchar buffer[256];

  switch(type) {
  case CHANGE_POINT :
    strcpy(buffer, GET_ENTRY(cur_point));
    sscanf(buffer, "%d", &size);
    break;
  case CHANGE_KANJI :
    strcpy(buffer, GET_ENTRY(cur_font[KANJI_FONT]));
    break;
  case CHANGE_ASCII :
    strcpy(buffer, GET_ENTRY(cur_font[ASCII_FONT]));
    break;
  case CHANGE_ULINE :
    state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(uline));
    flag = WordChr_UL;
    break;
  case CHANGE_BOLD :
    state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bold));
    flag = WordChr_BOLD;
    break;
  case CHANGE_ITALIC :
    state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(italic));
    flag = WordChr_IT;
    break;
  }
  start->y = end->y = -1;
  max.y = get_end_page();
  for(c.y = 0;c.y <= max.y;c.y++) {
    max.x = get_end_line(c.y);
    for(c.x = 0;c.x <=  max.x;c.x++) {
      if(master[c.y][c.x].attr & WordChr_REV) {
        switch(type) {
        case CHANGE_POINT :
        case CHANGE_KANJI :
        case CHANGE_ASCII :
          if(type == CHANGE_POINT) master[c.y][c.x].point = size;
          if(type == CHANGE_KANJI)
            master[c.y][c.x].findex = get_font_index(buffer, KANJI_FONT);
          if(type == CHANGE_ASCII)
            master[c.y][c.x].findex = get_font_index(buffer, ASCII_FONT);
          font = get_fontset(master[c.y][c.x]);
          if((master[c.y][c.x].code & 0xff00) != 0xff00)
            master[c.y][c.x].width = gdk_text_width_wc(font, &master[c.y][c.x].code, 1);
          else
            master[c.y][c.x].width = 0;
          break;
        case CHANGE_ULINE :
        case CHANGE_BOLD :
        case CHANGE_ITALIC :
          master[c.y][c.x].attr &= ~flag;
          if(state) master[c.y][c.x].attr |= flag;
          break;
        }
        end->y = c.y, end->x = c.x;
        if(start->y == -1) {
          start->y = c.y, start->x = c.x;
        }
      }
    }
  }
  if(start->y == -1) return;
  if(end->x < get_end_line(end->y)) {
    end->x++;
  } else if(end->y < get_end_page()){
    end->x = 0;
    end->y++;
  }
}

Position delete_master(Position start, Position end) {
  gint save, ex, cx, cy, incline, width, min, max, bx;

  incline = start.y - end.y;
  cy = end.y, cx = end.x;
  editbuf_pos = 0;
  while(master[cy][cx].code != GDK_Return) {
    editbuf[editbuf_pos++] = master[cy][cx];
    editbuf[editbuf_pos - 1].attr &= ~WordChr_CONT;
    if(master[cy][cx].attr & WordChr_CONT) {
      cy++;
      cx = -1;
      incline--;
    }
    cx++;
  }
  editbuf[editbuf_pos++] = master[cy][cx];
  editbuf[editbuf_pos - 1].attr &= ~WordChr_CONT;
  bx = width = 0;
  while(bx < start.x) width += master[start.y][bx++].width;
  cy = start.y;
  save = width, ex = cx = 0;
  while(ex < editbuf_pos) {
    if(editbuf[ex].code == GDK_Return || 
       ((width + editbuf[ex].width) >= (page_x - lines[cy].righttab - get_lefttab(cy)) && (cx + bx) > 0)) {
      incline++;
      cy++;
      cx = bx = width = 0;
    }
    width += editbuf[ex++].width;
    cx++;
  }
  incline--;
  while(incline < 0) {
    delete_line(start.y + 1);
    incline++;
  }
  while(incline > 0) {
    insert_line(start.y + 1);
    incline--;
  }
  ex = 0, cy = start.y, cx = start.x, width = save;
  while(ex < editbuf_pos) {
    if(width + editbuf[ex].width >= (page_x - lines[cy].righttab - get_lefttab(cy)) && cx > 0) {
        master[cy][cx - 1].attr |= WordChr_CONT;
        draw_line_pixmap(cy);
        width = 0;
    	cy++, cx = 0;
    }
    width += editbuf[ex].width;
    master[cy][cx++] = editbuf[ex++];
    if(master[cy][cx - 1].code == GDK_Return) {
      draw_line_pixmap(cy);
      width = 0;
      cy++, cx = 0;
    }
  }
  editbuf_pos = 0;
  set_lines_attr(start.y);
  get_lines_info(start.y, &min, &max);
  ajust_master(min, max);
  return start;
}

void insert_line(gint row) {
  gint i, s, y, max, cur, attr;
  GdkPixmap *pix;

  max = cur = get_end_page();
  y = row;
  if(cur >= MAX_ROW - 1) {
    cur = MAX_ROW - 2;
    lines[cur].attr |= LineAttr_EOF;
    if(lines[cur + 1].pixmap != NULL) {
      gdk_pixmap_unref(lines[cur + 1].pixmap);
      lines[cur + 1].pixmap = (GdkPixmap*)NULL;
    }
  }
  if(lines[cur + 1].pixmap != NULL) {
    gdk_pixmap_unref(lines[cur + 1].pixmap);
    lines[cur + 1].pixmap = NULL;
  }
  while(cur >= y) {
    lines[cur + 1] = lines[cur];
    lines[cur].attr = 0;
    lines[cur].pixmap = NULL;
    s = get_end_line(cur);
    for(i = 0;i <= s;i++) master[cur + 1][i] = master[cur][i];
    cur--;
  }
  if(y > max) {
    lines[y] = lines[y - 1];
    lines[y].attr |= LineAttr_NP | LineAttr_EOF;
    lines[y].pixmap = NULL;
    lines[y - 1].attr &= ~(LineAttr_NP | LineAttr_EOF);
  } else {
    if(y > 0) {
      attr = lines[y].attr;
      pix = lines[y].pixmap;
      lines[y] = lines[y - 1];
      lines[y].pixmap = pix;
      lines[y].attr |= attr & (LineAttr_NP | LineAttr_EOF);
    }
  }
}

void delete_line(gint row) {
  gint y, max, attr, s, i, j;

  if(row == 0) return;
  max = get_end_page();
  y = row;
  if(y > max) y = max;
  if(y >= max) {
    lines[y - 1].attr |= LineAttr_NP | LineAttr_EOF;
    return;
  }
  if(lines[y].pixmap != NULL) gdk_pixmap_unref(lines[y].pixmap);
  attr = lines[y].attr;
  for(i = y;i < max;i++) {
    lines[i] = lines[i + 1];
    s = get_end_line(i + 1);
    for(j = 0;j <= s;j++) master[i][j] = master[i + 1][j];
  }
  lines[i].pixmap = (GdkPixmap*)NULL;
  lines[i].attr = 0;
  lines[y].attr |= attr & (LineAttr_NP | LineAttr_EOF);
}

gint get_font_index(gchar *fontname, gint type) {
  gint i, index;

  index = 0;
  for(i = 0;fontnames[type][i][0] != '0';i++) {
    if(strcmp(fontname, fontnames[type][i]) == 0) {
      index = i;
      break;
    }
  }
  return index;
}

GdkFont* get_fontset(WordChr wc) {
  GdkFont *f;
  gint sa;
  gchar name[256];

  if((wc.code & 0xff00) == 0xff00) return NULL;
  if(wc.code > 0xff) {
    sa = (wc.attr & WordChr_IT) ? KANJI_ITALIC_FONT : KANJI_FONT;
  } else {
    sa = (wc.attr & WordChr_IT) ? ASCII_ITALIC_FONT : ASCII_FONT;
  }
  f = Fontarea[sa][wc.findex][wc.point];
  if(f != (GdkFont*)NULL) return f;
  if(wc.code > 0xff) {
    if(wc.attr & WordChr_IT)
      sprintf(name, "%si-normal-*-%d-*-*-*-*-*-jisx0208.1983-0,*",
	      xfontnames[KANJI_FONT][wc.findex], wc.point);
    else
      sprintf(name, "%sr-normal-*-%d-*-*-*-*-*-jisx0208.1983-0,*",
	      xfontnames[KANJI_FONT][wc.findex], wc.point);
  } else {
    if(wc.attr & WordChr_IT)
      sprintf(name, "%si-normal-*-%d-*-*-*-*-*-iso8859-*,*",
	      xfontnames[ASCII_FONT][wc.findex], wc.point);
    else
      sprintf(name, "%sr-normal-*-%d-*-*-*-*-*-iso8859-*,*",
	      xfontnames[ASCII_FONT][wc.findex], wc.point);
  }
  f = gdk_fontset_load(name);
  if(f == (GdkFont*)NULL) f = gdk_fontset_load("*");
  Fontarea[sa][wc.findex][wc.point] = f;
  return f;
}

gint get_line_height(gint y) {
  gint i, h;

  i = h = 0;
  do {
    if(h < master[y][i].point) h = master[y][i].point;
    i++;
  } while(end_line_query(y, i) == FALSE && i < MAX_COLUMN - 1);
  return h;
}

gint get_end_page() {
  gint i = 0;

  while(!(lines[i].attr & LineAttr_EOF) && (i < MAX_ROW - 1)) i++;
  return i;
}

gint get_end_line(gint y) {
  gint i = 0;

  while(end_line_query(y, i) == FALSE && i < MAX_COLUMN - 1) i++;
  return i;
}

gint end_line_query(gint y, gint x) {
  if(master[y][x].code == GDK_Return) return TRUE;
  if(master[y][x].attr & WordChr_CONT) return TRUE;
  return FALSE;
}

void set_center_aline() {
  Position start, end;
  gint i;

  set_rev(&start, &end, NONE);
  if(start.y == -1) {
    lines[master_.y].attr &= ~LineAttr_Form;
    lines[master_.y].attr |= LineAttr_Center;
    set_lines_attr(master_.y);
  } else {
    for(i = start.y;i <= end.y;i++) {
      lines[i].attr &= ~LineAttr_Form;
      lines[i].attr |= LineAttr_Center;
      set_lines_attr(i);
    }
  }
  draw_etext();
}

void set_left_aline() {
  Position start, end;
  gint i;

  set_rev(&start, &end, NONE);
  if(start.y == -1) {
    lines[master_.y].attr &= ~LineAttr_Form;
    lines[master_.y].attr |= LineAttr_Left;
    set_lines_attr(master_.y);
  } else {
    for(i = start.y;i <= end.y;i++) {
      lines[i].attr &= ~LineAttr_Form;
      lines[i].attr |= LineAttr_Left;
      set_lines_attr(i);
    }
  }
  draw_etext();
}

void set_right_aline() {
  Position start, end;
  gint i;

  set_rev(&start, &end, NONE);
  if(start.y == -1) {
    lines[master_.y].attr &= ~LineAttr_Form;
    lines[master_.y].attr |= LineAttr_Right;
    set_lines_attr(master_.y);
  } else {
    for(i = start.y;i <= end.y;i++) {
      lines[i].attr &= ~LineAttr_Form;
      lines[i].attr |= LineAttr_Right;
      set_lines_attr(i);
    }
  }
  draw_etext();
}

gint configure_event(GtkWidget *widget, GdkEventConfigure *event) {
  static gint first = 0;

  if(first == 0) {
    draw_line_pixmap(0);
    create_input_dialog(widget);
    first = 1;
  }
  return FALSE;
}

gint expose_event (GtkWidget *widget, GdkEventExpose *event) {
  draw_etext();
  return FALSE;
}

gint key_press_event (GtkWidget *widget, GdkEventKey *event) {
  gchar text[MAX_COLUMN];
  gint size, charcode;
  Position pos;
  GdkWChar cur_line[MAX_COLUMN], key;

  gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
  key = event->keyval;
  switch(key) {
  case GDK_Up :
    if(master_.y > 0) master_.y--;
    if(master_.x >= get_end_line(master_.y)) master_.x = get_end_line(master_.y);
    if(master_.y < disptop) draw_etext();
    set_word_combo();
    break;
  case GDK_Down :
    if(master_.y < get_end_page()) master_.y++;
    if(master_.x >= get_end_line(master_.y)) master_.x = get_end_line(master_.y);
    if((get_pix_y(master_.y) + get_line_height(master_.y)) >= page_y) set_page(cpage + 1);
    set_word_combo();
    break;
  case GDK_Left :
    if(master_.x > 0) {
      master_.x--;
    } else if(master_.y > 0){
      master_.y--;
      master_.x = get_end_line(master_.y);
      if(master_.y < disptop) set_page(cpage - 1);
    }
   set_word_combo();
   break;
  case GDK_BackSpace :
    if(master_.x == 0 && master_.y == 0) break;
    if(master_.x > 0) {
      master_.x--;
    } else if(master_.y > 0){
      master_.y--;
      master_.x = get_end_line(master_.y);
      if(master_.y < disptop) set_page(cpage - 1);
    }
  case GDK_Delete :
    pos.y = master_.y;
    pos.x = master_.x;
    if(pos.x < get_end_line(pos.y)) {
      pos.x++;
    } else if(pos.y < get_end_page()){
      pos.x = 0;
      pos.y++;
    }
    master_ = delete_master(master_, pos);
    draw_etext();
    set_word_combo();
    break;
  case GDK_Right :
    if(master_.x < get_end_line(master_.y)) {
      master_.x++;
    } else if(master_.y < get_end_page()){
      master_.x = 0;
      master_.y++;
    }
    if((get_pix_y(master_.y) + get_line_height(master_.y)) >= page_y) set_page(cpage + 1);
    set_word_combo();
    break;
  case GDK_Return:
    set_default_attr();
    editbuf_new();
    cur_line[0] = GDK_Return;
    cur_line[1] = 0;
    set_wcs_editbuf(cur_line);
    master_ = insert_editbuf(master_);
    draw_etext();
    break;
  default:
    /* ǥȥåפʤɤ٥Ȥ
       gtkΥ졼åȤ˽Ϥ褦ˡš */
    if(key != GDK_VoidSymbol && (key >= 0x80 || key < 0x20))
      return FALSE;
    if (event->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK))
      return FALSE;
  }
  if(key == GDK_VoidSymbol || (key < 0x80 && key >= 0x20)) {
    if(event->length > 0) {
      charcode = *(event->string);
      if (event->length == 1 && (charcode >= 0x80 || charcode < 0x20))
	return FALSE;
      strncpy(text, event->string, event->length);
      text[event->length] = '\0';
      ajust_mbtext(text);
      size = gdk_mbstowcs (cur_line, text, MAX_COLUMN);
      cur_line[size] = 0;
      set_default_attr();
      editbuf_new();
      set_wcs_editbuf(cur_line);
      master_ = insert_editbuf(master_);
      draw_etext();
    } else {
      return FALSE;
    }
  }
  tab_draw(tab_set, 0, TAB_DRAW);
  editbuf_new();
  set_input_dialog();
  return TRUE;
}

gint button_press_event (GtkWidget *widget, GdkEventButton *event) {
  static gint x, y, px, py;
  gint cx, cy, mx, my, redraw;

  y = get_word_y(event->y);
  x = get_word_x(y, event->x);
  if(event->button == 1) {
    if(event->type == GDK_BUTTON_PRESS) {
      master_.x = x;
      master_.y = y;
      px = event->x;
      py = event->y;
      tab_draw(tab_set, 0, TAB_DRAW);
      set_word_combo();
      set_input_dialog();
    } else if(event->type == GDK_BUTTON_RELEASE) {
      my = get_end_page();
      for(cy = 0;cy <= my;cy++) {
      	redraw = 0;
        mx = get_end_line(cy);
        for(cx = 0;cx <= mx;cx++) {
          if(master[cy][cx].attr & WordChr_REV) redraw = 1;
          master[cy][cx].attr &= ~WordChr_REV;
        }
        if(redraw == 1) draw_line_pixmap(cy);
      }
      if(px != event->x || py != event->y) {
        my = get_end_page();
        for(cy = master_.y;cy <= my && cy <= y;cy++) {
          mx = get_end_line(cy);
          cx = (cy == master_.y) ? master_.x : 0;
          while(cx <= mx && cx <= ((cy == y) ? x : mx)) {
            master[cy][cx].attr |= WordChr_REV;
            cx++;
          }
          draw_line_pixmap(cy);
        }
      }
      draw_etext();
    }
    return TRUE;
  }
  return FALSE;
}

gint motion_notify_event (GtkWidget *widget, GdkEventButton *event) {
  return FALSE;
}

gint leave_notify_event (GtkWidget *widget, GdkEventButton *event) {
  return FALSE;
}

gint enter_notify_event (GtkWidget *widget, GdkEventButton *event) {
  if(gdk_im_ready()) {
    gdk_im_begin(ic, etext->window);
  }
  tab_draw(tab_set, 0, TAB_END);
  gtk_widget_grab_focus(etext);
  return FALSE;
}

gint toggle_button_press(GtkWidget *widget, GdkEventButton *event, gint code) {
  gboolean status;

  gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_press_event");
  status = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), status ? FALSE : TRUE);
  change_word_attr(widget, code);
  return TRUE;
}

gint replace_tab(GtkWidget *widget) {
  tab_draw(widget, 0, TAB_DRAW);
  return FALSE;
}

gint tab_motion_notify(GtkWidget *widget, GdkEventMotion *event) {
  int x, y;
  GdkModifierType state;

  if(event->is_hint)
    gdk_window_get_pointer (event->window, &x, &y, &state);
  else {
    x = event->x;
    y = event->y;
    state = event->state;
  }
  if (state & GDK_BUTTON1_MASK)
    tab_draw(widget, x, TAB_SET);
  return TRUE;
}

gint tab_button_event(GtkWidget *widget, GdkEventButton *event) {
  gint x, y;

  x = event->x;
  y = event->y;
  if(event->button == 1) {
    if(event->type == GDK_BUTTON_PRESS) {
      tab_draw(widget, x, TAB_START);
    } else if(event->type == GDK_BUTTON_RELEASE) {
      tab_draw(widget, 0, TAB_END);
    }
  }
  return TRUE;
}

gint tab_draw(GtkWidget *widget, gint x, gint op) {
  static gint state = 0;
  Position start, end;
  gint   min, max;
  GdkGC  *white, *black, *bg;

  white = widget->style->white_gc;
  black = widget->style->black_gc;
  bg = widget->style->bg_gc[GTK_STATE_NORMAL];
  switch(op) {
  case TAB_END:
    if(state > 0) {
      set_rev(&start, &end, NONE);
      if(start.y == -1) {
        get_lines_info(master_.y, &min, &max);
        ajust_master(min, max);
      } else {
        get_lines_info(start.y, &min, &max);
        while(min <= end.y) {
          ajust_master(min, max);
          set_rev(&start, &end, NONE);
          if(get_lines_info(max + 1, &min, &max) == FALSE) break;
        }
      }
      draw_etext();
    }
    state = 0;
    break;
  case TAB_START:
    if(x > get_lefttab(master_.y) - 3 && x < get_lefttab(master_.y) + 3) state = 1;
    if(x > page_x - lines[master_.y].righttab - 3 && x < page_x - lines[master_.y].righttab + 3) state = 2;
  case TAB_SET:
    switch(state) {
    case 0:
      return FALSE;
    case 1:
      set_rev(&start, &end, NONE);
      if(start.y == -1) {
        lines[master_.y].lefttab = x;
        if(x < 0) lines[master_.y].lefttab = 0;
        if(x > page_x - lines[master_.y].righttab - 6)
          lines[master_.y].lefttab = page_x - lines[master_.y].righttab - 6;
        set_lines_attr(master_.y);
      } else {
        get_lines_info(start.y, &min, &max);
        while(min <= end.y) {
          lines[min].lefttab = x;
          if(x < 0) lines[min].lefttab = 0;
          if(x > page_x - lines[min].righttab - 6)
            lines[min].lefttab = page_x - lines[min].righttab - 6;
          set_lines_attr(min);
          if(get_lines_info(max + 1, &min, &max) == FALSE) break;
        }
      }
      break;
    case 2:
      set_rev(&start, &end, NONE);
      if(start.y == -1) {
        lines[master_.y].righttab = page_x - x;
        if(x < get_lefttab(master_.y) + 6)
          lines[master_.y].righttab = page_x - (get_lefttab(master_.y) + 6);
        if(x > page_x) lines[master_.y].righttab = 0;
        set_lines_attr(master_.y);
      } else {
        get_lines_info(start.y, &min, &max);
        while(min <= end.y) {
          lines[min].righttab = page_x - x;
          if(x < get_lefttab(min) + 6)
            lines[min].righttab = page_x - (get_lefttab(min) + 6);
          if(x > page_x) lines[min].righttab = 0;
          set_lines_attr(min);
          if(get_lines_info(max + 1, &min, &max) == FALSE) break;
        }
      }
      break;
    }
  case TAB_DRAW:
    gdk_draw_rectangle(widget->window, bg, TRUE, 0, 0, A3HEIGHT, 16);
    gdk_draw_line(widget->window, black, 0, 7, page_x, 7);
    gdk_draw_line(widget->window, white, 0, 10, page_x, 10);
    gdk_draw_rectangle(widget->window, bg, TRUE, get_lefttab(master_.y) - 4, 3, 9, 8);
    gdk_draw_line(widget->window, white, get_lefttab(master_.y) - 4, 3, get_lefttab(master_.y) + 4, 3);
    gdk_draw_line(widget->window, white, get_lefttab(master_.y) - 4, 3, get_lefttab(master_.y) - 4, 12);
    gdk_draw_line(widget->window, black, get_lefttab(master_.y) - 4, 12, get_lefttab(master_.y) + 4, 12);
    gdk_draw_line(widget->window, black, get_lefttab(master_.y) + 4, 3, get_lefttab(master_.y) + 4, 12);
    gdk_draw_rectangle(widget->window, bg, TRUE, page_x - lines[master_.y].righttab - 4, 3, 9, 8);
    gdk_draw_line(widget->window, white, page_x - lines[master_.y].righttab - 4, 3, page_x - lines[master_.y].righttab + 4, 3);
    gdk_draw_line(widget->window, white, page_x - lines[master_.y].righttab - 4, 3, page_x - lines[master_.y].righttab - 4, 12);
    gdk_draw_line(widget->window, black, page_x - lines[master_.y].righttab - 4, 12, page_x - lines[master_.y].righttab + 4, 12);
    gdk_draw_line(widget->window, black, page_x - lines[master_.y].righttab + 4, 3, page_x - lines[master_.y].righttab + 4, 12);
    break;
  case TAB_MAX:
    lines[master_.y].righttab = page_x - x;
    set_lines_attr(master_.y);
    break;
  }
  return FALSE;
}

void ajust_mbtext(gchar *text) {
  static gchar mblists[] = "áʡˡΡϡСѡܡ";
  static gchar clists[] = ",.:;?!'`\"^~_/|()[]{}+=<>\\$%#&*@";
  gint src, dst, p, max, cmax;

  src = dst = 0;
  max = strlen(text);
  cmax = strlen(clists);
  while(src < max) {
    if(!(text[src] & 0x80)) {
      text[dst++] = text[src++];
    } else {
      for(p = 0;p < cmax;p++) {
        if(text[src] == mblists[p * 2] && text[src + 1] == mblists[p * 2 + 1]) {
          text[dst++] = clists[p];
          src += 2;
          break;
        }
      }
      if(p == cmax) {
      	if(text[src] == (gchar)0xa3 && text[src + 1] >= (gchar)0xb0 && text[src + 1] <= (gchar)0xb9) {
          text[dst++] = (text[src + 1] - 0xb0) + '0';
          src += 2;
      	} else if(text[src] == (gchar)0xa3 && text[src + 1] >= (gchar)0xc1 && text[src + 1] <= (gchar)0xda) {
          text[dst++] = (text[src + 1] - 0xc1) + 'A';
          src += 2;
     	} else if(text[src] == (gchar)0xa3 && text[src + 1] >= (gchar)0xe1 && text[src + 1] <= (gchar)0xfa) {
          text[dst++] = (text[src + 1] - 0xe1) + 'a';
          src += 2;
      	} else {
          text[dst++] = text[src++];
          text[dst++] = text[src++];
        }
      }
    }
  }
  text[dst] = '\0';
}

gint change_word_attr(GtkWidget *widget, gint code) {
  Position start, end;
  gint min, max;

  if(change_combo-- > 0) return TRUE;
  set_rev(&start, &end, NONE);
  if(start.y == -1) return FALSE;
  switch(code) {
  case CHANGE_POINT :
  case CHANGE_KANJI :
  case CHANGE_ASCII :
  case CHANGE_ULINE :
  case CHANGE_BOLD :
  case CHANGE_ITALIC :
    set_rev(&start, &end, code);
    get_lines_info(start.y, &min, &max);
    while(min <= end.y) {
      ajust_master(min, max);
      set_rev(&start, &end, NONE);
      if(get_lines_info(max + 1, &min, &max) == FALSE) break;
    }
    break;
  }
  set_page(cpage);
  draw_etext();
  return TRUE;
}

void set_word_combo() {
  WordChr wc;
  gchar buf[10];

  wc = master[master_.y][master_.x];
  if(wc.code == GDK_Return) {
    if(master_.x == 0) return;
    wc = master[master_.y][master_.x - 1];
  }
  if((wc.code & 0xff00) == 0xff00) return;
  sprintf(buf, "%d", wc.point);
  change_combo = 2;
  gtk_entry_set_text(POINT_ENTRY, buf);
  change_combo = 2;
  if(wc.code > 0xff) {
    gtk_entry_set_text(KANJI_ENTRY, fontnames[KANJI_FONT][wc.findex]);
  } else {
    gtk_entry_set_text(ASCII_ENTRY, fontnames[ASCII_FONT][wc.findex]);
  }
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(uline), (wc.attr & WordChr_UL) ? TRUE : FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bold), (wc.attr & WordChr_BOLD) ? TRUE : FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(italic), (wc.attr & WordChr_IT) ? TRUE : FALSE);
}

void set_tabular() {
  gchar buffer[32];

  sprintf(buffer, "1");
  number_dialog("Ȥο(164)", buffer, (GtkSignalFunc)tabular_ok);
}

void tabular_ok(GtkWidget *widget, GtkWidget **widgets) {
  GtkWidget *dialog, *entry;
  gint i, n, step, length, min, max, mx;

  dialog = widgets[0];
  entry = widgets[1];
  sscanf(gtk_entry_get_text(GTK_ENTRY(entry)), "%d", &n);
  if(n < 1 || n > 32) return;
  if(lines[master_.y].attr & LineAttr_Tabular != 0) {
    gtk_widget_destroy(dialog);
    return;
  }
  step = page_x / n;
  length = 0;
  get_lines_info(master_.y, &min, &max);
  max = min - 1;
  for(i = 1;i <= n;i++) {
    if(max + 1 > get_end_page()) {
      insert_line(max + 1);
      mx = get_end_line(max);
      master[max + 1][0] = master[max][mx];
      master[max + 1][0].code = GDK_Return;
    }
    get_lines_info(max + 1, &min, &max);
    lines[min].lefttab = length;
    length += step;
    lines[min].righttab = page_x - length;
    if(i > 1) lines[min].attr |= LineAttr_Tabular;
    set_lines_attr(min);
    ajust_master(min, max);
    get_lines_info(min, &min, &max);
  }
  draw_etext();
  gtk_widget_destroy(dialog);
}

void set_lines_attr(gint y) {
  gint i, cy, mx, my, min, max, attr;
  LineAttr line;
  GdkPixmap *pixmap;
  /* Position range = {-1, -1}; */

  my = get_end_page();
  if(y > my) return;
  cy = y;
  do {
    if(--cy < 0) break;
    mx = get_end_line(cy);
  } while(master[cy][mx].code != GDK_Return);
  min = cy + 1;
  do {
    mx = get_end_line(++cy);
  } while(master[cy][mx].code != GDK_Return);
  max = cy;
  line = lines[y];
  for(i = min;i <= max;i++) {
    pixmap = lines[i].pixmap;
    attr = lines[i].attr;
    lines[i] = line;
    lines[i].pixmap = pixmap;
    lines[i].attr &= ~(LineAttr_NP | LineAttr_EOF);
    lines[i].attr |= attr & (LineAttr_NP | LineAttr_EOF);
  }
}

gint get_lines_info(gint y, gint *min, gint *max) {
  gint cy, mx, my;

  my = get_end_page();
  if(y > my) return FALSE;
  cy = y;
  do {
    if(--cy < 0) break;
    mx = get_end_line(cy);
  } while(master[cy][mx].code != GDK_Return);
  *min = cy + 1;
  do {
    mx = get_end_line(++cy);
  } while(master[cy][mx].code != GDK_Return);
  *max = cy;
  return TRUE;
}

void print_word_destroy(GtkWidget *widget) {
  gtk_grab_remove(widget);
}


/*****************************************
 * ʸΰˤĤƤγѿ
 *****************************************/
void init_document_area () {
  gint i, j, k;
  gchar *cur;

  for (i = 0; i < 3; i++)
    Memory[i] = g_malloc (MAX_ROW * MAX_COLUMN * sizeof(WordChr));

  /* Ϥʬmaster[] */
  cur = Memory[0];
  for(i = 0;i < MAX_ROW;i++) {
    master[i] = (WordChr*)cur;
    cur += MAX_COLUMN * sizeof(WordChr);
  }

  /* ʬcopybuf[] */
  cur = Memory[1];
  for(i = 0;i < MAX_ROW;i++) {
    copybuf[i] = (WordChr*)cur;
    cur += MAX_COLUMN * sizeof(WordChr);
  }

  /* ʬeditbuf */
  cur = Memory[2];
  editbuf = (WordChr*)cur;
  cur += (MAX_ROW * MAX_COLUMN * sizeof(WordChr)) / 2;

  /* Ǹʬadjustbuf */
  ajustbuf = (WordChr*)cur;

  /* 򥯥ꥢ */
  for(i = 0;i < MAX_ROW;i++) lines[i].pixmap = (GdkPixmap*)NULL;

  /* ν */
  master_.x = master_.y = 0;

  /* ܤ° */
  lines[0].attr = LineAttr_NP | LineAttr_EOF;
  lines[0].vspace = 0;
  lines[0].lefttab = 0;
  lines[0].righttab = 0;
  lines[0].firsttab = 0;
  lines[0].reserved1 = 0;
  lines[0].reserved2 = 0;
  lines[0].reserved3 = 0;
  lines[0].reserved4 = 0;
  set_default_attr();
  master[0][0].attr = editbuf_attr;
  master[0][0].point = editbuf_point;
  master[0][0].findex = editbuf_findex[ASCII_FONT];
  master[0][0].code = GDK_Return;
  master[0][0].width = 0;

  /* ǥեȤΥեȤƤ */
  for(i = 0;i < 4;i++)
    for(j = 0;j < MAX_FONT;j++)
      for(k = 0;k < 256;k++) Fontarea[i][j][k] = (GdkFont*)NULL;
  gdk_mbstowcs (&(master[0][0].code), "", 2);
  master[0][0].findex = editbuf_findex[KANJI_FONT];
  get_fontset (master[0][0]);
  gdk_mbstowcs (&(master[0][0].code), " ", 1);
  master[0][0].findex = editbuf_findex[ASCII_FONT];
  get_fontset (master[0][0]);
  master[0][0].code = GDK_Return;

  page_x = paper_width - paper_left - paper_right;
  page_y = paper_height - paper_top - paper_bottom;
  max_dsp_num = page_y / 7;
  pix_width = page_x * 2;
  pix_height = 256;
  gtk_timeout_add(50, (GtkFunction)blink_cursor, (gpointer)NULL);
  copy_.y = -1;
  cpage = disptop = 0;
}

/*****************************************
 * λΥХå
 *****************************************/
void word_main_quit() {
  int i;

  for(i = 0;i < 3;i++)
    g_free (Memory[i]);

  gtk_main_quit();
}

/*****************************************
 * Ԥyɸ׻
 *****************************************/
gint inc_line_pix_y(gint type, inc_pixy *value) {
  gint height;

  switch(type) {
  case INC_ZERO:
    value->current = 0;
    value->y = 0;
    break;
  case INC_INIT:
    value->current = 0;
    value->y = disptop;
    break;
  case INC_COUNT:
    height = get_line_height(value->y);
    value->current += lines[value->y].vspace + height + height / 6;
    (value->y)++;
    break;
  }
  return value->current;
}

/*****************************************
 * Ԥyɸ
 *****************************************/
gint get_pix_y(gint y) {
  gint pix_y, i;
  inc_pixy value;

  pix_y = inc_line_pix_y(INC_INIT, &value);
  for(i = disptop;i < y && !(lines[i].attr & LineAttr_EOF) && (i < MAX_ROW - 1);i++)
    pix_y = inc_line_pix_y(INC_COUNT, &value);
  return pix_y;
}

/*****************************************
 * Խʸ°뤫
 *****************************************/
void set_default_attr() {
  sscanf(GET_ENTRY(cur_point), "%d", &editbuf_point);
  editbuf_findex[KANJI_FONT] = get_font_index(GET_ENTRY(CUR_KANJI), KANJI_FONT);
  editbuf_findex[ASCII_FONT] = get_font_index(GET_ENTRY(CUR_ASCII), ASCII_FONT);
  editbuf_attr = 0;
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(uline))) editbuf_attr |= WordChr_UL;
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bold))) editbuf_attr |= WordChr_BOLD;
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(italic))) editbuf_attr |= WordChr_IT;
}

/*****************************************
 * ԽХåե˴
 *****************************************/
void editbuf_new() {
  editbuf_pos = 0;
}

/*****************************************
 * Ϳ줿wcsʸ°ͿԽХåե
 *****************************************/
void set_wcs_editbuf(GdkWChar *wcs) {
  gint i = 0;
  GdkFont *font;

  while(wcs[i] != 0) {
    editbuf[editbuf_pos].code = wcs[i];
    editbuf[editbuf_pos].point = editbuf_point;
    editbuf[editbuf_pos].attr = editbuf_attr;
    editbuf[editbuf_pos].findex = editbuf_findex[(wcs[i] < 0x100) ? ASCII_FONT : KANJI_FONT];
    if((wcs[i] & 0xff00) == 0xff00) {
      if(wcs[i] != GDK_Return) {
      	i++;
      	continue;
      }	
      editbuf[editbuf_pos].width = 0;
    }
    else {
      font = get_fontset(editbuf[editbuf_pos]);
      editbuf[editbuf_pos].width = gdk_text_width_wc(font, &editbuf[editbuf_pos].code, 1);
    }
    editbuf_pos++;
    i++;
  }
}

/*****************************************
 * ǤκǾʹֹ
 *****************************************/
gint get_page_top(gint pg, gint *cp) {
  return get_page_info(pg, cp, PAGEINFO_NUM);
}

/*****************************************
 * ǽǤκǾʹֹ
 *****************************************/
gint get_endpage_top(gint *cp) {
  return get_page_info(1, cp, PAGEINFO_MAX);
}

/*****************************************
 * Ǥξ
 *****************************************/
gint get_page_info(gint pg, gint *cp, gint flag) {
  gint i, page, height, cur, l, save, w;
  inc_pixy value;

  *cp = 0;
  l = 0;
  if(pg <= 0) return 0;
  save = inc_line_pix_y(INC_ZERO, &value);
  height = page = 0;
  for(i = 0;!(lines[i].attr & LineAttr_EOF) && (i < MAX_ROW - 1);i++) {
    w = inc_line_pix_y(INC_COUNT, &value);
    cur = w - save;
    save = w;
    height += cur;
    if(height >= page_y) {
      page++;
      l = i;
      *cp = page;
      if(pg == page && flag == PAGEINFO_NUM) return l;
      height = cur;
    }
  }
  return l;
}
